﻿using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using System.Windows.Media.Imaging;
using System.Collections.Generic;

namespace SeriTile
{
    public class Entity: Observable
    {
        public enum Direction { North, South, West, East, NorthEast, NorthWest, SouthEast, SouthWest };
        List<SeriTile.IntPoint> m_path = null;
        private TileMap m_map;
        
        public Sprite Sprite {get; protected set;}
        public Point m_worldLocation;
        public IntPoint m_tileLocation;
        public Point Target;
        public Direction Facing { get; set; }
        public float Speed;
        public bool Moving {get; private set;}
        public TileMap.TileInfo m_curTile;

        public Point WorldLocation
        {
            get
            {
                return m_worldLocation;
            }
            set
            {
                m_worldLocation = value;
                m_tileLocation = m_map.worldToTilePosition(m_worldLocation);
                if (m_curTile != null)
                    m_curTile.Entities.Remove(this);
                m_curTile = m_map.GetTileInfo(m_tileLocation.X, m_tileLocation.Y);
                m_curTile.Entities.Add(this);
            }
        }

        public IntPoint TileLocation
        {
            get
            {
                return m_tileLocation;
            }
            set
            {
                m_tileLocation = value;
                if (m_map != null)
                {
                    if (m_map.Type == TileMap.MapType.Orthogonal)
                    {

                        m_worldLocation = m_tileLocation.ToPoint(m_map.TileWidth);
                    }
                    else
                    {
                        m_worldLocation = m_map.IsometricToWorld(m_tileLocation);
                    }
                    if (m_curTile != null)
                        m_curTile.Entities.Remove(this);
                    m_curTile = m_map.GetTileInfo(m_tileLocation.X, m_tileLocation.Y);
                    m_curTile.Entities.Add(this);
                }
            }
        }

        public TileMap Map 
        {
            get
            {
                return m_map;
            }
            set
            {
                if (m_map != value)
                {
                    if(m_path != null)
                        m_path.Clear();
                    StopMoving();
                }
                m_map = value;
                if (m_curTile != null)
                    m_curTile.Entities.Remove(this);
                m_curTile = null;
            }
        }

        public Entity(Sprite sprite)
        {
            Sprite = sprite;
            Moving = false;
            Facing = Direction.South;
            Speed = 2;
            StopMoving();
        }

        public Entity()
        {
            Sprite = null;
            Moving = false;
            Facing = Direction.South;
            Speed = 2;
            StopMoving();
        }

        public Point GetBlitPoint()
        {
            Point blitPoint = new Point();
            blitPoint.X = WorldLocation.X + Sprite.Tile.Offset.X - Map.DrawOffset.X;
            blitPoint.Y = WorldLocation.Y + Sprite.Tile.Offset.Y - Map.DrawOffset.Y;
            return blitPoint;
        }

        public bool IsClicked(Point topLeft, int tileWidth, int tileHeight, Point clickPoint)
        {
            Point blitPoint = GetBlitPoint();

            System.Windows.Rect rect = new Rect(blitPoint.X, blitPoint.Y, Sprite.Tile.Rect.Width, Sprite.Tile.Rect.Height);
            if (rect.Contains(clickPoint))
            {
                Point pixelPos = new Point(clickPoint.X - rect.X, clickPoint.Y - rect.Y);
                uint pixel = (uint)Sprite.Tile.Image.Pixels[(int)(pixelPos.X + pixelPos.Y * Sprite.Tile.Image.PixelWidth)];
                uint alpha = pixel >> 12;
                return alpha == 0 ? false : true;
            }
            return false;
        }

        public virtual void Draw(WriteableBitmap writeableBmp)
        {
            Point blitPoint = GetBlitPoint();
            Sprite.Tile.Blit(writeableBmp, blitPoint);
        }

        public void UpdateCamera(KeyHandler keyHandler, float deltaTime, int screenWidth, int screenHeight)
        {

            double x, y;

            x = WorldLocation.X - (screenWidth / 2.0f) - m_map.TileWidth / 2;
            if (x < m_map.MinOffset.X)
                x = m_map.MinOffset.X;
            if (x > ((int)m_map.MaxOffset.X))
                x = ((int)m_map.MaxOffset.X);
       

            y = WorldLocation.Y - (screenHeight / 2.0f) - m_map.TileHeight / 2;
            if (y < m_map.MinOffset.Y)
                y = m_map.MinOffset.Y;
            if (y > ((int)m_map.MaxOffset.Y))
                y = ((int)m_map.MaxOffset.Y);
            m_map.DrawOffset = new Point(x,y);
        }

        public void MoveTo(IntPoint position)
        {
            //If we are moving path from our current target, if we are still path from our location
            if (m_path!= null && m_path.Count > 0)
            {
                IntPoint curTarget = m_path[m_path.Count - 1];
                AStar.GetSingleton().GetPath(Map, curTarget.X, curTarget.Y, position.X, position.Y, out m_path, false);
            }
            else
            {
                AStar.GetSingleton().GetPath(Map, TileLocation.X, TileLocation.Y, position.X, position.Y, out m_path, false);
            }
            if(!Moving)
                Moving = true;
        }

        //Moves to the tile next to the position
        public void MoveNextTo(IntPoint position)
        {
            //If we are moving path from our current target, if we are still path from our location
            if (m_path != null && m_path.Count > 0)
            {
                IntPoint curTarget = m_path[m_path.Count - 1];
                AStar.GetSingleton().GetPath(Map, curTarget.X, curTarget.Y, position.X, position.Y, out m_path, true);
            }
            else
            {
                AStar.GetSingleton().GetPath(Map, TileLocation.X, TileLocation.Y, position.X, position.Y, out m_path, true);
            }
            if(!Moving)
                Moving = true;
        }

        public void MovePath(List<IntPoint> path)
        {
            m_path = path;
            Moving = true;
        }

        public virtual void Update(float deltaTime)
        {
            if (m_path != null && m_path.Count > 0)
            {
                SeriTile.IntPoint Pathpoint = m_path[m_path.Count - 1];
                Point pathLoc;
                if (m_map.Type == TileMap.MapType.Orthogonal)
                {
                    pathLoc = Pathpoint.ToPoint(m_map.TileWidth);
                }
                else
                {
                    pathLoc = m_map.IsometricToWorld(Pathpoint);
                }
                Point dir = new Point(pathLoc.X - WorldLocation.X, pathLoc.Y - WorldLocation.Y);
                double distance = Math.Sqrt(dir.X * dir.X + dir.Y * dir.Y);
                if (distance == 0)
                {
                    m_path.RemoveAt(m_path.Count - 1);
                }
                else
                {

                    // Normalize dir
                    dir.X = dir.X / distance;
                    dir.Y = dir.Y / distance;

                    double moveDist = deltaTime * Speed;
                    moveDist *= m_map.TileHeight;
                    if (moveDist > distance)
                    {
                        moveDist = distance;

                        // Remove the point since we've reached it
                        m_path.RemoveAt(m_path.Count - 1);

                        //Call an event
                        IntPoint point = TileLocation;
                        SeriTile.TileMap.TileInfo tileInfo = Map.GetTileInfo(point.X, point.Y);
                        if (tileInfo.TileEvent != null)
                            tileInfo.TileEvent(tileInfo, new SeriTile.TileMap.TileEventArgs(0, 0, this));
                    }
                    if (moveDist > 0 && Moving)
                    {
                        WorldLocation = new Point(WorldLocation.X + dir.X * moveDist, WorldLocation.Y + dir.Y * moveDist);

                        if(dir.Y < 0)
                        {
                            if(dir.X < 0)
                            {
                                Sprite.changeAnimation("WalkNorthWest");//UpLeft
                                Facing = Direction.NorthWest;
                            }
                            else if(dir.X > 0)
                            {
                                Sprite.changeAnimation("WalkNorthEast");//UpRight
                                Facing = Direction.NorthEast;
                            }
                            else
                            {
                                Sprite.changeAnimation("WalkNorth");
                                Facing = Direction.North;
                            }
                        }
                        else if(dir.Y > 0)
                        {
                            if(dir.X < 0)
                            {
                                Sprite.changeAnimation("WalkSouthWest");//DownLeft
                                Facing = Direction.SouthWest;
                            }
                            else if(dir.X > 0)
                            {
                                Sprite.changeAnimation("WalkSouthEast");//DownRight
                                Facing = Direction.SouthEast;
                            }
                            else
                            {
                                Sprite.changeAnimation("WalkSouth");
                                Facing = Direction.South;
                            }
                        }
                        else if(dir.X < 0)
                        {
                            Sprite.changeAnimation("WalkWest");
                            Facing = Direction.West;
                        }
                        else if (dir.X > 0)
                        {
                            Sprite.changeAnimation("WalkEast");
                            Facing = Direction.East;
                        }
                    }
                }
            }
            else
            {
                if (Moving)
                    StopMoving();
            }
            Sprite.update(deltaTime);
        }

        public void PlayAnimation(string animation)
        {
            Sprite.changeAnimation(animation);
        }

        public void StopMoving()
        {
            Moving = false;
            m_path = null;
            if(Sprite != null)
            {
                switch (Facing)
                {
                    case Entity.Direction.North:
                        Sprite.changeAnimation("StandNorth");
                        break;
                    case Entity.Direction.South:
                        Sprite.changeAnimation("StandSouth");
                        break;
                    case Entity.Direction.West:
                        Sprite.changeAnimation("StandWest");
                        break;
                    case Entity.Direction.East:
                        Sprite.changeAnimation("StandEast");
                        break;
                    case Entity.Direction.NorthEast:
                        Sprite.changeAnimation("StandNorthEast");
                        break;
                    case Entity.Direction.SouthWest:
                        Sprite.changeAnimation("StandSouthWest");
                        break;
                    case Entity.Direction.NorthWest:
                        Sprite.changeAnimation("StandNorthWest");
                        break;
                    case Entity.Direction.SouthEast:
                        Sprite.changeAnimation("StandSouthEast");
                        break;
                }
            }
        }
    }
}
