﻿using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using System.Xml.Linq;
using System.IO;
using System.Windows.Media.Imaging;
using System.Collections.Generic;

namespace SeriTile
{
    //23
    public class TileMap
    {
        public enum MapType { Orthogonal, IsometricStaggered, IsometricDiamond };
        public MapType Type{get; private set;}

        public class TileEventArgs:EventArgs
        {
            public TileEventArgs(int x, int y, Entity entity)
            {
                this.x = x;
                this.y = y;
                this.entity = entity;
            }
            int x;
            int y;
            Entity entity;
        }
        public delegate void TileEventHandler(TileInfo sender, EventArgs TileEventArgs);
        public TileEventHandler OnTMXEvent;

        public class TileInfo
        {
            public TileInfo(int layerCount)
            {
                Tile = new ITile[layerCount];
                Entities = new List<Entity>();
            }
            public bool Collision { get; set; }
            public ITile[] Tile {get; set;}
            public TileEventHandler TileEvent;
            public List<Entity> Entities { get; private set; }
            public TileTag tag;
        }

        private TileInfo[,] m_tileInfo;

        public int Width { get; private set;}
        public int Height { get; private set; }
        public int LayerCount { get; private set; }
        public int EntityLayer { get; private set; }//Entities are drawn just after this layer.
        public int TileWidth { get; private set; }
        public int TileHeight { get; private set; }
        public Point DrawOffset { get; set; }
        public Point MinOffset { get; set; }
        public Point MaxOffset { get; set; }
        private System.Collections.Generic.Dictionary<string, TileTag> m_tags = new System.Collections.Generic.Dictionary<string, TileTag>();
        public static ITile PickHelper { get; set; }
        //Width and height in pixels of the map
        public int PixelWidth { get; private set; }
        public int PixelHeight { get; private set; }

        public TileMap(MapType type, int xSize, int ySize, int tileWidth, int tileHeight, int layerCount, int entityLayer, WriteableBitmap screen)
        {
            Type = type;
            m_tileInfo = new TileInfo[xSize, ySize];
            EntityLayer = entityLayer;
            TileWidth = tileWidth;
            TileHeight = tileHeight;
            Width = xSize;
            Height = ySize;
            LayerCount = layerCount;

            if(screen == null)
                SetBounds(0,0);
            else
                SetBounds(screen.PixelWidth, screen.PixelWidth);

            for (int i = 0; i < xSize; i++)
            {
                for (int j = 0; j < ySize; j++)
                {
                    m_tileInfo[i, j] = new TileInfo(layerCount);
                }
            }
        }

        void SetBounds(int screenWidth, int screenHeight)
        {
            if (Type == MapType.IsometricStaggered)
            {
                PixelWidth = TileWidth * Width;
                PixelHeight = TileHeight * Height / 2;
                MinOffset = new Point(TileWidth / 2, TileHeight / 2);
                MaxOffset = new Point(PixelWidth - screenWidth, PixelHeight - screenHeight);
            }
            else
            {
                if (Type == MapType.Orthogonal)
                {
                    PixelWidth = TileWidth * Width;
                    PixelHeight = TileHeight * Height;
                    MinOffset = new Point();
                    MaxOffset = new Point(PixelWidth - screenWidth, PixelHeight - screenHeight);
                }
                else//Diagonal Iso
                {
                    PixelWidth = TileWidth * Width;
                    PixelHeight = TileHeight * Height;
                    MinOffset = new Point(-PixelWidth / 2, 0);
                    MaxOffset = new Point(PixelWidth / 2 + TileWidth - screenWidth, (PixelWidth - screenHeight)/ 2);
                }
            }
        }

        /// <summary>
        /// Loads a map from a TileEd tmx file
        /// </summary>
        /// <param name="resourcePath"></param>
        /// <param name="fileName"></param>
        public TileMap(string resourcePath, string fileName, TileSetManager tilesetManager, WriteableBitmap screen)
        {
            XDocument doc = XDocument.Load(resourcePath + fileName);

            XElement mapElem = doc.Root;
            string type = (string)mapElem.Attribute("orientation");
            if (type == "orthogonal")
                Type = MapType.Orthogonal;
            else
                Type = MapType.IsometricDiamond;
            int tileWidth = int.Parse(mapElem.Attribute("tilewidth").Value);
            int tileHeight = int.Parse(mapElem.Attribute("tileheight").Value);
            TileWidth = tileWidth;
            TileHeight = tileHeight;
            Width = int.Parse(mapElem.Attribute("width").Value);
            Height = int.Parse(mapElem.Attribute("height").Value);

            SetBounds(screen.PixelWidth, screen.PixelHeight);

            XElement mapProperties = doc.Root.Element("properties");
            XElement mapProperty = mapProperties.Element("property");
            EntityLayer = (int)mapProperty.Attribute("value");

            LayerCount = 0;
            foreach (XElement layerElem in mapElem.Elements("layer"))
            {
                bool isCollision = false;
                XElement layerProps = layerElem.Element("properties");
                if(layerProps != null)
                {
                    foreach (XElement layerProp in layerProps.Elements("property"))
                    {
                        if (((string)layerProp.Attribute("Type")) == "Collision")
                            isCollision = true;
                    }
                }
                if (!isCollision)
                    LayerCount++;
            }

            m_tileInfo = new TileInfo[Width, Height];
            for (int i = 0; i < Width; i++)
            {
                for (int j = 0; j < Height; j++)
                {
                    m_tileInfo[i, j] = new TileInfo(LayerCount);
                }
            }

            System.Collections.Generic.List<TileSet> m_tileSets = new System.Collections.Generic.List<TileSet>();
            foreach (XElement tileSet in doc.Root.Elements("tileset"))
            {
                string filename = (string)tileSet.Attribute("source");
                int gid = (int)tileSet.Attribute("firstgid");
                if (filename != null)//Load a tileset file
                {
                    TileSet tileset = tilesetManager.LoadTileSet(resourcePath+filename);
                    tileset.GID = gid;
                    m_tileSets.Add(tileset);
                }
                else//Load a raw tileset
                {
                    string name = (string)tileSet.Attribute("name");
                    XElement imageElem = tileSet.Element("image");
                    string sourceFile = (string)imageElem.Attribute("source");
                    int baseHeight;
                    if (Type == MapType.IsometricStaggered || Type == MapType.IsometricDiamond)
                        baseHeight = TileHeight;
                    else
                        baseHeight = TileWidth;

                    int tilesetTileWidth = int.Parse(tileSet.Attribute("tilewidth").Value);
                    int tilesetTileHeight = int.Parse(tileSet.Attribute("tileheight").Value);

                    TileSet tileset = tilesetManager.LoadTileSet(name, resourcePath + sourceFile, tilesetTileWidth, tilesetTileHeight, baseHeight, WriteableBitmapExtensions.BlendMode.Alpha);
                    tileset.GID = gid;
                    m_tileSets.Add(tileset);
                }
            }

            int layerNum = 0;
            foreach (XElement layerElem in doc.Root.Elements("layer"))
            {
                if (layerNum == 6)
                {
                    layerNum++;
                    continue;
                }
                bool isCollision = false;
                XElement data = layerElem.Element("data");
                string encoding = data.Attribute("encoding").Value;
                if (encoding == "base64")
                {
                    int dataSize = (Width * Height * 4) + 1024;
                    string value = data.Value;
                    byte[] buffer = Convert.FromBase64String(value);

                    Stream stream = new MemoryStream(buffer, true);
                    using (stream)
                    using (var br = new BinaryReader(stream))
                    {

                        XElement properties = layerElem.Element("properties");
                        if (properties != null)
                        {
                            foreach (XElement propElem in properties.Elements("property"))
                            {
                                string name = propElem.Attribute("name").Value;
                                string val = propElem.Attribute("value").Value;
                                if (name == "Type" && val == "Collision")
                                    isCollision = true;
                            }
                        }
                        if (isCollision)
                        {
                            for (int i = 0; i < Height; i++)
                            {
                                for (int j = 0; j < Width; j++)
                                {
                                    int n = br.ReadInt32();
                                    TileInfo tileInfo = GetTileInfo(j,i);
                                    if(tileInfo != null)
                                        tileInfo.Collision = (n == 0 ? true : false);
                                }
                            }
                            layerNum--;
                        }
                        else
                        {
                            ITile tile = null;
                            for (int i = 0; i < Height; i++)
                            {
                                for (int j = 0; j < Width; j++)
                                {
                                    Int32 tileIndex = br.ReadInt32();
                                    foreach (TileSet tileset in m_tileSets)
                                    {
                                        if (tileIndex <= tileset.MaxGID)
                                        {
                                            tile = tileset.getTileFromGID(tileIndex);
                                            break;
                                        }
                                    }
                                    TileInfo tileInfo = GetTileInfo(j,i);
                                    if(tileInfo != null)
                                        tileInfo.Tile[layerNum] = tile;
                                }
                            }
                        }
                    }
                    layerNum++;
                }
            }

            XElement objectGroup = doc.Root.Element("objectgroup");
            foreach (XElement oj in objectGroup.Elements("object"))
            {
                TileTag tag = null;
                if(Type == MapType.Orthogonal)
                    tag = makeTag(oj, tileWidth, tileHeight);
                else
                    tag = makeTag(oj, tileHeight, tileHeight);
                m_tags.Add(tag.Name, tag);
                TileInfo tileInfo = GetTileInfo(tag.Point.X, tag.Point.Y);
                if (tileInfo != null)
                {
                    tileInfo.tag = tag;
                    if (oj.Attribute("type").Value == "Event")
                    {
                        tileInfo.TileEvent += new TileEventHandler(GenericTileEvent);
                    }
                }

            }
            

        }

        private TileTag makeTag(XElement oj, int tileWidth, int tileHeight)
        {
            int x = (int)oj.Attribute("x");
            int y = (int)oj.Attribute("y");
            int px = x / tileWidth;
            int py = y / tileHeight;

            TileTag tag = new TileTag();
            tag.Name = (string)oj.Attribute("name");
            XElement properties = oj.Element("properties");


            tag.Point = new IntPoint(px, py);
            if (properties != null)
            {
                foreach (XElement property in properties.Elements("property"))
                {
                    string dataName = (string)property.Attribute("name");
                    string data = (string)property.Attribute("value");
                    tag.Data.Add(dataName, data);
                }
            }

            return tag;
        }

        /// <summary>
        /// Gets the clicked square.
        /// If iso, PickHelper needs to be set up to a tile that looks like the red/yellow/green/blue tile at
        /// the bottom of http://www.gamedev.net/reference/articles/article747.asp
        /// </summary>
        /// <param name="clickPos"></param>
        /// <param name="offset"></param>
        /// <returns></returns>
        public IntPoint getSquareUnderScreenPoint(Point clickPos)
        {
            return worldToTilePosition(new Point(clickPos.X + DrawOffset.X, clickPos.Y + DrawOffset.Y));
        }

        /// <summary>
        /// Converts a world position to a tile point.
        /// </summary>
        /// <param name="clickPos"></param>
        /// <param name="offset"></param>
        /// <returns></returns>
        public IntPoint worldToTilePosition(Point clickPos)
        {
            SeriTile.IntPoint clickLoc = null;
            if (Type == SeriTile.TileMap.MapType.Orthogonal)
            {
                clickLoc = new SeriTile.IntPoint((int)(clickPos.X / TileWidth), (int)(clickPos.Y / TileHeight));
            }
            else if (Type == SeriTile.TileMap.MapType.IsometricStaggered)
            {
                int x = (int)((clickPos.X) / TileWidth);
                int y = (int)((clickPos.Y) / TileHeight) * 2;
                int rx = (int)((clickPos.X) % TileWidth + PickHelper.Rect.X);
                int ry = (int)((clickPos.Y) % TileHeight + PickHelper.Rect.Y);
                uint pixel = (uint)PickHelper.Image.Pixels[(int)(rx + ry * PickHelper.Image.PixelWidth)];
                uint alpha = (pixel >> 24) & 0xFF;
                uint red = (pixel >> 16) & 0xFF;
                uint green = (pixel >> 8) & 0xFF;
                uint blue = (pixel) & 0xFF;
                if (alpha == 0)
                {
                    clickLoc = new SeriTile.IntPoint(x, y);
                }
                else if (red > 0)
                {
                    if (green > 0)//Yellow
                    {
                        clickLoc = new SeriTile.IntPoint(x, y - 1);
                    }
                    else//Red
                    {
                        clickLoc = new SeriTile.IntPoint(x - 1, y - 1);
                    }
                }
                else if (green > 0)//Green
                {
                    clickLoc = new SeriTile.IntPoint(x - 1, y + 1);
                }
                else if (blue > 0)//Blue
                {
                    clickLoc = new SeriTile.IntPoint(x, y + 1);
                }
            }
            else
            {
                int x = (int)Math.Round(clickPos.X / TileWidth + clickPos.Y / TileHeight) - 1;
                int y = (int)Math.Round(-clickPos.X / TileWidth + clickPos.Y / TileHeight);
                clickLoc = new IntPoint(x, y);
            }
            return clickLoc;
        }

        public TileTag getTag(string tag)
        {
            if (m_tags.ContainsKey(tag))
            {
                return m_tags[tag];
            }
            return null;
        }

        public virtual bool CanWalk(SeriTile.Entity ignoreEntity, IntPoint a, IntPoint b)
        {
            if (a.X < 0 || a.Y < 0 || a.X >= Width || a.Y >= Height)
                return false;
            if (b.X < 0 || b.Y < 0 || b.X >= Width || b.Y >= Height)
                return false;
            return m_tileInfo[b.X, b.Y].Collision;
        }

        public virtual int GetCost(IntPoint a, IntPoint b)
        {
            if (CanWalk(null, a, b))
                return 1;
            else
                return 0;
        }

        public virtual void getConnectingPoints(Entity entity, IntPoint pos, out List<IntPoint> pointList)
        {
            pointList = new List<IntPoint>();
            if (Type == MapType.Orthogonal || Type == MapType.IsometricDiamond)
            {
                IntPoint point = new IntPoint(pos.X, pos.Y + 1);
                if (CanWalk(entity, pos, point))
                    pointList.Add(point);
                point = new IntPoint(pos.X, pos.Y - 1);
                if (CanWalk(entity, pos, point))
                    pointList.Add(point);
                point = new IntPoint(pos.X + 1, pos.Y);
                if (CanWalk(entity, pos, point))
                    pointList.Add(point);
                point = new IntPoint(pos.X - 1, pos.Y);
                if (CanWalk(entity, pos, point))
                    pointList.Add(point);
                point = new IntPoint(pos.X + 1, pos.Y + 1);
                if (CanWalk(entity, pos, point))// && CanWalk(pos, new IntPoint(pos.X, pos.Y + 1)) && CanWalk(pos, new IntPoint(pos.X + 1, pos.Y)))
                    pointList.Add(point);
                point = new IntPoint(pos.X - 1, pos.Y + 1);
                if (CanWalk(entity, pos, point))// && CanWalk(pos, new IntPoint(pos.X, pos.Y + 1)) && CanWalk(pos, new IntPoint(pos.X - 1, pos.Y)))
                    pointList.Add(point);
                point = new IntPoint(pos.X + 1, pos.Y - 1);
                if (CanWalk(entity, pos, point))// && CanWalk(pos, new IntPoint(pos.X, pos.Y - 1)) && CanWalk(pos, new IntPoint(pos.X + 1, pos.Y)))
                    pointList.Add(point);
                point = new IntPoint(pos.X - 1, pos.Y - 1);
                if (CanWalk(entity, pos, point))// && CanWalk(pos, new IntPoint(pos.X, pos.Y - 1)) && CanWalk(pos, new IntPoint(pos.X - 1, pos.Y)))
                    pointList.Add(point);
            }
            else
            {
                IntPoint point = new IntPoint(pos.X + 1, pos.Y);//East
                if (CanWalk(entity, pos, point))
                    pointList.Add(point);
                point = new IntPoint(pos.X - 1, pos.Y);//West
                if (CanWalk(entity, pos, point))
                    pointList.Add(point);
                point = new IntPoint(pos.X, pos.Y - 2);//North
                if (CanWalk(entity, pos, point))
                    pointList.Add(point);
                point = new IntPoint(pos.X, pos.Y + 2);//South
                if (CanWalk(entity, pos, point))
                    pointList.Add(point);
                point = new IntPoint(pos.X + (pos.Y % 2 == 0 ? 0 : 1), pos.Y + 1);//SouthEast
                if (CanWalk(entity, pos, point) && CanWalk(entity, pos, new IntPoint(pos.X, pos.Y + 2)) && CanWalk(entity, pos, new IntPoint(pos.X + 1, pos.Y)))
                    pointList.Add(point);
                point = new IntPoint(pos.X - (pos.Y % 2 == 0 ? 1 : 0), pos.Y + 1);//SouthWest
                if (CanWalk(entity, pos, point) && CanWalk(entity, pos, new IntPoint(pos.X, pos.Y + 2)) && CanWalk(entity, pos, new IntPoint(pos.X - 1, pos.Y)))
                    pointList.Add(point);
                point = new IntPoint(pos.X + (pos.Y % 2 == 0 ? 0 : 1), pos.Y - 1);//NorthEast
                if (CanWalk(entity, pos, point) && CanWalk(entity, pos, new IntPoint(pos.X, pos.Y - 2)) && CanWalk(entity, pos, new IntPoint(pos.X + 1, pos.Y)))
                    pointList.Add(point);
                point = new IntPoint(pos.X - (pos.Y % 2 == 0 ? 1 : 0), pos.Y - 1);//NorthWest
                if (CanWalk(entity, pos, point) && CanWalk(entity, pos, new IntPoint(pos.X, pos.Y - 2)) && CanWalk(entity, pos, new IntPoint(pos.X - 1, pos.Y)))
                    pointList.Add(point);
            }
        }
        /*
        public Point IsometricToWorld(Point isometricCoord)
        {
            Point worldCoord = new Point();
            worldCoord.X = isometricCoord.X * TileWidth + ((int)(isometricCoord.Y) & 1) * (TileWidth / 2) - TileWidth / 2;
            worldCoord.Y = isometricCoord.Y * TileHeight / 2;
            return worldCoord;
        }*/

        /// <summary>
        /// Returns the world coordinates of the CENTER of the given tile
        /// </summary>
        /// <param name="isometricCoord"></param>
        /// <returns></returns>
        public Point IsometricToWorld(IntPoint isometricCoord)
        {
            Point worldCoord = new Point();
            if (Type == MapType.IsometricStaggered)
            {
                worldCoord.X = isometricCoord.X * TileWidth + (isometricCoord.Y % 2 == 1 ? TileWidth / 2 : 0) + TileWidth / 2;
                worldCoord.Y = isometricCoord.Y * (TileHeight / 2) + TileHeight / 2;
            }
            else
            {
                worldCoord.X = (isometricCoord.X * TileWidth / 2) - (TileWidth / 2 * isometricCoord.Y) + TileWidth/2;
                worldCoord.Y = (isometricCoord.Y * TileHeight / 2) + (TileHeight / 2 * isometricCoord.X) + TileHeight/2;
            }
            return worldCoord;
        }

        private void DrawIso(WriteableBitmap writeableBmp)
        {
            Point blitPoint = new Point();
            int tileWidth = writeableBmp.PixelWidth / TileWidth;
            int tileHeight = writeableBmp.PixelHeight / TileHeight;


            if (Type == MapType.IsometricStaggered)
            {
                for (int k = 0; k < LayerCount; k++)
                {
                    //offset.X = topLeftMap.X;
                    //offset.Y = topLeftMap.Y;
                    for (int j = 0, b = 0; b < Height; j += TileHeight / 2, b++)
                    {
                        for (int i = 0, a = 0; a < Width; i += TileWidth, a++)
                        {
                            double offsetX = (b % 2 == 1 ? -TileWidth / 2 : 0);
                            TileInfo tileInfo = GetTileInfo(a, b);
                            if (tileInfo != null)
                            {
                                ITile tile = tileInfo.Tile[k];
                                if (tile != null)
                                {
                                    double offsetY = tile.Rect.Height - TileHeight;
                                    blitPoint.X = i - offsetX - DrawOffset.X;
                                    blitPoint.Y = j - offsetY - DrawOffset.Y;
                                    tile.Blit(writeableBmp, blitPoint);
                                }
                                if (k == EntityLayer)
                                {
                                    foreach (Entity entity in tileInfo.Entities)
                                    {
                                        if (entity != null)
                                        {
                                            entity.Draw(writeableBmp);
                                        }
                                    }
                                }
                            }
                        }
                    }


                }
            }
            else
            {
                Point drawPoint;
                for (int k = 0; k < LayerCount; k++)
                {
                    for (int b = 0; b < Height; b++)
                    {
                        drawPoint = IsometricToWorld(new IntPoint(0, b));
                        drawPoint.X -= TileWidth / 2;
                        drawPoint.Y -= TileHeight / 2;
                        for (int a = 0; a < Height; a++)
                        {

                            TileInfo tileInfo = GetTileInfo(a, b);
                            if (tileInfo != null)
                            {
                                ITile tile = tileInfo.Tile[k];
                                if (tile != null)
                                {


                                    double offsetY = tile.Rect.Height - TileHeight;
                                    blitPoint.X = drawPoint.X - DrawOffset.X;
                                    blitPoint.Y = drawPoint.Y - offsetY - DrawOffset.Y;
                                    tile.Blit(writeableBmp, blitPoint);
                                }
                                if (k == EntityLayer)
                                {
                                    foreach (Entity entity in tileInfo.Entities)
                                    {
                                        if (entity != null)
                                        {
                                            entity.Draw(writeableBmp);
                                        }
                                    }
                                }
                            }
                            drawPoint.X += TileWidth / 2;
                            drawPoint.Y += TileHeight / 2;
                        }
                    }
                }
            }
        }


        private void DrawOrtho(WriteableBitmap writeableBmp)
        {
            Point blitPoint = new Point();
            int tileCountX = writeableBmp.PixelWidth / TileWidth;
            int tileCountY = writeableBmp.PixelHeight / TileHeight;
            Point offset = new Point(DrawOffset.X % TileWidth, DrawOffset.Y % TileHeight);

            for (int k = 0; k < LayerCount; k++)
            {
                IntPoint tilePos = new IntPoint(DrawOffset, TileWidth);
                for (int j = 0, b = tilePos.Y; b < tilePos.Y + tileCountY + 2; j += TileHeight, b++)
                {
                    for (int i = 0, a = tilePos.X; a < tilePos.X + tileCountX + 1; i += TileWidth, a++)
                    {
                        TileInfo tileInfo = GetTileInfo(a, b);
                        if (tileInfo != null)
                        {
                            ITile tile = tileInfo.Tile[k];
                            if (tile != null)
                            {
                                blitPoint.X = i - offset.X;
                                blitPoint.Y = j - tile.Offset.Y - offset.Y;
                                //Debug coloring
                                //Color testColor = tileInfo.Entity != null ? Colors.Yellow : (!tileInfo.Collision ? Colors.Red : Colors.White);
                                //tile.Blit(writeableBmp, blitPoint, testColor);
                                tile.Blit(writeableBmp, blitPoint);
                            }

                            if (k == EntityLayer)
                            {
                                foreach (Entity entity in tileInfo.Entities)
                                {
                                    if (entity != null)
                                    {
                                        entity.Draw(writeableBmp);
                                    }
                                }
                            }
                        }
                    }
                }


            }
        }

        public virtual void Draw(WriteableBitmap writeableBmp)
        {
            if (Type == MapType.Orthogonal)
                DrawOrtho(writeableBmp);
            else
                DrawIso(writeableBmp);
        }

        public void GenericTileEvent(TileInfo sender, EventArgs TileEventArgs)
        {
            if(OnTMXEvent != null)
                OnTMXEvent(sender, TileEventArgs);
        }

        public TileInfo GetTileInfo(int x, int y)
        {
            if (x < 0 || y < 0 || x >= Width || y >= Height)
                return null;
            return m_tileInfo[x, y];
        }

    }


}
