﻿using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using System.Collections.Generic;
using System.Windows.Media.Imaging;
using System.IO;
using System.Xml.Linq;
using System.Linq;

namespace SeriTile
{
    /// <summary>
    /// Need to add to loader:
    /// Does not need Width and Height on TileSet tag
    /// Width and height on tile tag
    /// Offset on tile tag
    /// Does not need frametime on AnimatedTile tag
    /// FrameTime on Frame tag
    /// </summary>
    public class TileSet
    {
        private System.Collections.Generic.List<ITile> m_tiles = new System.Collections.Generic.List<ITile>();
        private System.Collections.Generic.Dictionary<string, ITile> m_namedTiles = new System.Collections.Generic.Dictionary<string, ITile>();
        private Random m_rand = new Random();
        private int m_tileWidth;
        private int m_tileHeight;
        private System.Collections.Generic.List<ITile> m_animatedTiles = new System.Collections.Generic.List<ITile>();
        public int GID { get; set; }
        public int MaxGID
        {
            get
            {
                return GID + m_tiles.Count - 1;
            }
        }

        internal TileSet(string imageFile, int tileWidth, int tileHeight, int baseHeight, WriteableBitmapExtensions.BlendMode blendMode)
        {
            m_tileHeight = tileHeight;
            m_tileWidth = tileWidth;
            addTiles(imageFile, blendMode, baseHeight);
        }

        internal TileSet(string resourceRoot, string xmlDefinition)
        {
            XDocument doc = XDocument.Load(resourceRoot + xmlDefinition);
            if(doc.Root.Attribute("Height") != null)
                m_tileHeight = (int)doc.Root.Attribute("Height");
            if (doc.Root.Attribute("Width") != null)
                m_tileWidth = (int)doc.Root.Attribute("Width");
            string blendModeStr = (string)doc.Root.Attribute("BlendMode");
            WriteableBitmapExtensions.BlendMode defaultBlendMode = (WriteableBitmapExtensions.BlendMode)Enum.Parse(typeof(WriteableBitmapExtensions.BlendMode), blendModeStr, true);
            string fileName = resourceRoot + (string)doc.Root.Attribute("File");

            bool uniformType = true;
            if((string)doc.Root.Attribute("Type")!= null)
            {
                string type = (string)doc.Root.Attribute("Type");
                if(type.ToLower() == "uniform")
                    uniformType = true;
                else if(type.ToLower() == "nonuniform")
                    uniformType = false;
            }

            XAttribute baseHeightAttribute = doc.Root.Attribute("BaseHeight");
            int baseHeight = m_tileWidth;
            if (baseHeightAttribute != null)
                baseHeight = (int)baseHeightAttribute;

            WriteableBitmap image = null;
            int numTilesX = 0;
            if (uniformType)
            {
                addTiles(fileName, WriteableBitmapExtensions.BlendMode.Alpha, baseHeight);
                numTilesX = m_tiles[0].Image.PixelWidth / m_tileWidth;
            }
            else
                image = LoadBitmap(fileName);
             

            foreach (XElement tileElement in doc.Root.Elements("Tile"))
            {
                string name = (string)tileElement.Attribute("Name");
                int x = (int)tileElement.Attribute("X");
                int y = (int)tileElement.Attribute("Y");
                blendModeStr = (string)tileElement.Attribute("BlendMode");
                WriteableBitmapExtensions.BlendMode blendMode;
                if (blendModeStr == null)
                {
                    blendMode = defaultBlendMode;
                }
                else
                {
                    blendMode = (WriteableBitmapExtensions.BlendMode)Enum.Parse(typeof(WriteableBitmapExtensions.BlendMode), blendModeStr, true);
                }
                ITile tile = null;
                //Tile tile = new Tile(new Rect(x*m_tileWidth, y*m_tileHeight, m_tileWidth, m_tileHeight), offset, image, blendMode);
                if(uniformType)
                    tile = getTile(x + y * numTilesX + 1);
                else
                {
                    int width = (int)tileElement.Attribute("Width");
                    int height = (int)tileElement.Attribute("Height");
                    int offsetX = (int)tileElement.Attribute("XOffset");
                    int offsetY = (int)tileElement.Attribute("YOffset");
                    tile = new Tile(new Rect(x, y, width, height), new Point(offsetX, offsetY), image, blendMode);
                }

                tile.BlendMode = blendMode;
                m_namedTiles.Add(name, tile);
            }
            foreach (XElement animElement in doc.Root.Elements("AnimatedTile"))
            {
                string name = (string)animElement.Attribute("Name");
                float defaultFrameTime = (float)animElement.Attribute("FrameTime");

                bool loop = true;
                if (animElement.Attribute("Loop") != null)
                {
                    loop = (bool)animElement.Attribute("Loop");
                }

                AnimatedTile tile = new AnimatedTile(name, loop);
                foreach (XElement frameElement in animElement.Elements("Frame"))
                {
                    string tileName = (string)frameElement.Attribute("Tile");
                    ITile frameTile = getTile(tileName);

                    float frameTime = defaultFrameTime;
                    if (frameElement.Attribute("FrameTime") != null)
                        frameTime = (float)frameElement.Attribute("FrameTime");
                    if(frameTile != null)
                        tile.addFrame(frameTile, frameTime);
                }
                m_animatedTiles.Add(tile);
                m_tiles.Add(tile);
                m_namedTiles.Add(name, tile);
            }

        }

        private struct StringAnim
        {
            public StringAnim(string str)
            {
                animString = str;
                index = -1;
                anim = null;
            }
            public int index;
            public AnimatedTile anim;
            public string animString;
        }

        public TileSet(XDocument doc)
        {
            if (doc.Root.Name == "tileset")
            {
                LoadTileset(doc.Root);
            }
            else
            {
                foreach (XElement tileset in doc.Root.Elements("tileset"))
                {
                    LoadTileset(tileset);
                }
            }
        }

        private void LoadTileset(XElement tileset)
        {
            m_tileWidth = int.Parse(tileset.Attribute("tilewidth").Value);
            m_tileHeight = int.Parse(tileset.Attribute("tileheight").Value);
            Point offset = new Point(0, m_tileHeight - m_tileWidth);

            XElement imageElem = tileset.Element("image");
            string fileName = imageElem.Attribute("source").Value;
            WriteableBitmap image = LoadBitmap(@"/SilverlightTest;component/Resources/" + fileName);

            WriteableBitmapExtensions.BlendMode blendMode = WriteableBitmapExtensions.BlendMode.None;
            WriteableBitmapExtensions.BlendMode defaultBlendMode = WriteableBitmapExtensions.BlendMode.None;


            System.Collections.Generic.List<StringAnim> animations = new List<StringAnim>();
            int startIndex = m_tiles.Count;
            int n = 0;
            for (int i = 0; i + m_tileHeight <= image.PixelHeight; i += m_tileHeight)
            {
                for (int j = 0; j + m_tileWidth <= image.PixelWidth; j += m_tileWidth)
                {
                    IEnumerable<XElement> tileElems = from x in tileset.Elements("tile") where (int)x.Attribute("id") == n select x;
                    foreach (XElement tileElem in tileElems)
                    {
                        bool useDefaultBlend = true;
                        XElement properties = tileElem.Element("properties");
                        foreach (XElement property in properties.Elements("property"))
                        {
                            string name = property.Attribute("name").Value;
                            string value = property.Attribute("value").Value;

                            if (name == "BlendingAll")
                            {
                                defaultBlendMode = (WriteableBitmapExtensions.BlendMode)Enum.Parse(typeof(WriteableBitmapExtensions.BlendMode), value, true);
                            }
                            if (name == "Blending")
                            {
                                blendMode = (WriteableBitmapExtensions.BlendMode)Enum.Parse(typeof(WriteableBitmapExtensions.BlendMode), value, true);
                                useDefaultBlend = false;
                            }
                            if (name == "Animated")
                            {
                                animations.Add(new StringAnim(value));
                            }
                        }
                        if (useDefaultBlend)
                            blendMode = defaultBlendMode;
                    }

                    Tile tile = new Tile(new Rect(j, i, m_tileWidth, m_tileHeight), offset, image, blendMode);
                    m_tiles.Add(tile);
                    n++;
                }
            }

            for (int i = 0; i < animations.Count; i++)
            {
                StringAnim animString = animations[i];
                string[] indexes = animString.animString.Split(',');
                animString.anim = new AnimatedTile(i.ToString(), true);
                foreach (string indexString in indexes)
                {
                    int index = Int32.Parse(indexString);
                    if (animString.index == -1)
                        animString.index = index;
                    animString.anim.addFrame(m_tiles[startIndex + index],0.3f);
                }
                animations[i] = animString;
            }
            for (int i = 0; i < animations.Count; i++)
            {
                StringAnim animString = animations[i];
                m_tiles[startIndex + animString.index] = animString.anim;
                m_animatedTiles.Add(animString.anim);
            }
        }

        public void update(float deltaTime)
        {
            foreach (AnimatedTile tile in m_animatedTiles)
            {
                tile.update(deltaTime);
            }
        }

        public void addTiles(string filename, WriteableBitmapExtensions.BlendMode blendMode, int baseHeight)
        {
            WriteableBitmap image = LoadBitmap(filename);

            Point offset = new Point(0, m_tileHeight - baseHeight);
            for (int i = 0; i < image.PixelHeight; i += m_tileHeight)
            {
                for (int j = 0; j < image.PixelWidth; j += m_tileWidth)
                {
                    Tile tile = new Tile(new Rect(j, i, m_tileWidth, m_tileHeight), offset, image, blendMode);
                    m_tiles.Add(tile);
                }
            }
        }

        public ITile getRandomTile()
        {
            return m_tiles[m_rand.Next(m_tiles.Count)];
        }
        
        public ITile getTile(int i)
        {
            if (i == 0)
                return null;
            return m_tiles[i-1];
        }

        public ITile getTile(string i)
        {
            if (m_namedTiles.Keys.Contains(i))
            {
                return m_namedTiles[i];
            }
            return null;
        }

        public ITile getTileFromGID(int i)
        {
            if (i - GID < 0)
                return null;
            return m_tiles[i - GID];
        }


        private WriteableBitmap LoadBitmap(string path)
        {
            BitmapImage img = new BitmapImage();
            img.CreateOptions = BitmapCreateOptions.None;
            Stream s = Application.GetResourceStream(new Uri(path, UriKind.Relative)).Stream;
            img.SetSource(s);
            return new WriteableBitmap(img);
        }
    }
}
